from .color import Color
from . import colors
from . import csscolors
from . import res

from .cp437 import to_cp437

from .input import Keys

from .terminal import Terminal
from .terminal import TerminalBuilder

from .app import App

__all__ = [
    "Color",
    "colors",
    "csscolors",
    "App",
    "Keys",
    "to_cp437",
    "res",
    "Terminal",
    "TerminalBuilder",
]



from abc import abstractmethod
from pathlib import Path
from typing import Protocol


class Font(Protocol):
    _width: int
    _height: int

    _tile_width: int
    _tile_height: int

    _image_path: Path

    def __init__(self, tile_width: int, tile_height: int, image: Path):
        self._tile_width = tile_width
        self._tile_height = tile_height
        self._image_path = image

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, value: int) -> None:
        raise RuntimeError("Changing the size of a Font is not supported")

    @property
    def height(self) -> int:
        return self._height

    @height.setter
    def height(self, value: int) -> None:
        raise RuntimeError("Changing the size of a Font is not supported")

    @property
    def tile_width(self) -> int:
        return self._tile_width

    @tile_width.setter
    def tile_width(self, value: int) -> None:
        raise RuntimeError("Changing the size of a Font is not supported")

    @property
    def tile_height(self) -> int:
        return self._tile_height

    @tile_height.setter
    def tile_height(self, value: int) -> None:
        raise RuntimeError("Changing the size of a Font is not supported")

    @abstractmethod
    def load(self) -> None:
        raise NotImplementedError("Font loading not implemented")



from pathlib import Path

import arcade
from arcade.gl import Texture2D

from baggo.terminal import Font


class FontArcade(Font):
    _initialized: bool = False
    _image_path: Path

    _texture: Texture2D
    _columns: int
    _rows: int

    def __init__(self, tile_width: int, tile_height: int, image: Path):
        super().__init__(tile_width, tile_height, image)
        self._image_path = image

    def load(self) -> None:
        self._texture = arcade.get_window().ctx.load_texture(self._image_path)
        self._columns = self._texture.width // self._tile_width
        self._rows = self._texture.height // self._tile_height
        self._initialized = True

    @property
    def initialized(self):
        return self._initialized

    @property
    def texture(self) -> Texture2D:
        return self._texture

    @property
    def columns(self) -> int:
        return self._columns

    @property
    def rows(self) -> int:
        return self._rows



from baggo.terminal import Console, Terminal
from baggo.input import Keys


class App:
    terminal: Terminal
    console: Console

    def __init__(self, terminal: Terminal):
        self.terminal = terminal
        self.terminal.register_app(self)
        self.console = terminal.console

    def run(self) -> None:
        self.terminal.run()

    def tick(self, delta_time: float) -> None:
        pass

    def on_key_down(self, key: Keys, modifiers: int) -> None:
        pass

    def on_key_up(self, key: Keys, modifiers: int) -> None:
        pass



from .console import Console, SimpleConsole

from .font import Font
from .terminal import Terminal
from .terminal_builder import TerminalBuilder

__all__ = ["Console", "Font", "SimpleConsole", "Terminal", "TerminalBuilder"]



from abc import abstractmethod
from typing import Any, Protocol, TYPE_CHECKING

from .console import Console

if TYPE_CHECKING:
    from baggo import App


class Terminal(Protocol):
    _app: App
    _width: int
    _height: int

    console: Console

    @abstractmethod
    def register_app(self, app: App):
        raise NotImplementedError

    @abstractmethod
    def run(self) -> None:
        raise NotImplementedError

    @property
    def app(self) -> Any:
        return self._app

    @app.setter
    def app(self, value: App) -> None:
        raise RuntimeError("Changing the app of a Terminal is not supported")

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, value: int) -> None:
        raise RuntimeError("Changing the size of a Terminal is not supported")

    @property
    def height(self) -> int:
        return self._height

    @height.setter
    def height(self, value: int) -> None:
        raise RuntimeError("Changing the size of a Terminal is not supported")



from enum import Enum
from pathlib import Path

from .console import Console, SimpleConsole
from .font import Font
from .terminal import Terminal


class TerminalBackends(Enum):
    ARCADE = 1
    PYGLET = 2
    PYGAME = 3


DEFAULT_TERMINAL_BACKEND = TerminalBackends.ARCADE


class TerminalBuilder:
    backend: TerminalBackends

    width: int
    height: int

    title: str

    console: Console
    font: Font

    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

    @classmethod
    def simple(
        cls,
        width: int,
        height: int,
        title: str,
        tile_width: int,
        tile_height: int,
        font_file: Path,
        backend: TerminalBackends = DEFAULT_TERMINAL_BACKEND,
    ) -> TerminalBuilder:
        con = SimpleConsole(width, height)
        tb = cls(width, height)
        tb.title = title
        tb.console = con
        tb.backend = backend
        tb.font = TerminalBuilder.create_font(
            tb.backend, tile_width, tile_height, Path(font_file)
        )
        return tb

    @staticmethod
    def create_font(
        backend: TerminalBackends, tile_width: int, tile_height: int, font_file: Path
    ):
        match backend:
            case TerminalBackends.ARCADE:
                from .backends.arcade import FontArcade

                return FontArcade(tile_width, tile_height, font_file)

        raise RuntimeError(f"Unable to create font for current backend {backend.name}")

    def build(self) -> Terminal:
        match self.backend:
            case TerminalBackends.ARCADE:
                from .backends.arcade import FontArcade, TerminalArcade

                assert isinstance(self.font, FontArcade)
                return TerminalArcade(
                    self.width * self.font.tile_width,
                    self.height * self.font.tile_height,
                    self.title,
                    self.console,
                    self.font,
                )

        raise RuntimeError(
            f"Unable to create terminal for current backend {self.backend.name}"
        )



from abc import abstractmethod
from dataclasses import dataclass
from typing import Protocol

from baggo import colors, Color


@dataclass
class Tile:
    glyph: int
    foreground: Color
    background: Color


class Console(Protocol):
    width: int
    height: int

    dirty: bool

    @abstractmethod
    def at(self, x: int, y: int) -> Tile | None:
        raise NotImplementedError

    @abstractmethod
    def clear(self, color: Color = colors.BLACK) -> None:
        raise NotImplementedError

    @abstractmethod
    def print(
        self,
        x: int,
        y: int,
        text: str,
        foreground: Color | None = None,
        background: Color | None = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def set(
        self,
        x: int,
        y: int,
        glyph: int,
        foreground: Color,
        background: Color | None = None,
    ) -> None:
        raise NotImplementedError



from enum import Enum
from sys import platform


class Input(Enum):
    pass


class Keys(Input):
    # Key modifiers
    # Done in powers of two, so you can do a bit-wise 'and' to detect
    # multiple modifiers.
    MOD_SHIFT = 1
    MOD_CTRL = 2
    MOD_ALT = 4
    MOD_CAPSLOCK = 8
    MOD_NUMLOCK = 16
    MOD_WINDOWS = 32
    MOD_COMMAND = 64
    MOD_OPTION = 128
    MOD_SCROLLLOCK = 256

    # Platform-specific base hotkey modifier
    MOD_ACCEL = MOD_COMMAND if platform == "darwin" else MOD_CTRL

    # Keys
    BACKSPACE = 65288
    TAB = 65289
    LINEFEED = 65290
    CLEAR = 65291
    RETURN = 65293
    ENTER = 65293
    PAUSE = 65299
    SCROLLLOCK = 65300
    SYSREQ = 65301
    ESCAPE = 65307
    HOME = 65360
    LEFT = 65361
    UP = 65362
    RIGHT = 65363
    DOWN = 65364
    PAGEUP = 65365
    PAGEDOWN = 65366
    END = 65367
    BEGIN = 65368
    DELETE = 65535
    SELECT = 65376
    PRINT = 65377
    EXECUTE = 65378
    INSERT = 65379
    UNDO = 65381
    REDO = 65382
    MENU = 65383
    FIND = 65384
    CANCEL = 65385
    HELP = 65386
    BREAK = 65387
    MODESWITCH = 65406
    SCRIPTSWITCH = 65406
    MOTION_UP = 65362
    MOTION_RIGHT = 65363
    MOTION_DOWN = 65364
    MOTION_LEFT = 65361
    MOTION_NEXT_WORD = 1
    MOTION_PREVIOUS_WORD = 2
    MOTION_BEGINNING_OF_LINE = 3
    MOTION_END_OF_LINE = 4
    MOTION_NEXT_PAGE = 65366
    MOTION_PREVIOUS_PAGE = 65365
    MOTION_BEGINNING_OF_FILE = 5
    MOTION_END_OF_FILE = 6
    MOTION_BACKSPACE = 65288
    MOTION_DELETE = 65535
    NUMLOCK = 65407
    NUM_SPACE = 65408
    NUM_TAB = 65417
    NUM_ENTER = 65421
    NUM_F1 = 65425
    NUM_F2 = 65426
    NUM_F3 = 65427
    NUM_F4 = 65428
    NUM_HOME = 65429
    NUM_LEFT = 65430
    NUM_UP = 65431
    NUM_RIGHT = 65432
    NUM_DOWN = 65433
    NUM_PRIOR = 65434
    NUM_PAGE_UP = 65434
    NUM_NEXT = 65435
    NUM_PAGE_DOWN = 65435
    NUM_END = 65436
    NUM_BEGIN = 65437
    NUM_INSERT = 65438
    NUM_DELETE = 65439
    NUM_EQUAL = 65469
    NUM_MULTIPLY = 65450
    NUM_ADD = 65451
    NUM_SEPARATOR = 65452
    NUM_SUBTRACT = 65453
    NUM_DECIMAL = 65454
    NUM_DIVIDE = 65455

    # Numbers on the numberpad
    NUM_0 = 65456
    NUM_1 = 65457
    NUM_2 = 65458
    NUM_3 = 65459
    NUM_4 = 65460
    NUM_5 = 65461
    NUM_6 = 65462
    NUM_7 = 65463
    NUM_8 = 65464
    NUM_9 = 65465

    F1 = 65470
    F2 = 65471
    F3 = 65472
    F4 = 65473
    F5 = 65474
    F6 = 65475
    F7 = 65476
    F8 = 65477
    F9 = 65478
    F10 = 65479
    F11 = 65480
    F12 = 65481
    F13 = 65482
    F14 = 65483
    F15 = 65484
    F16 = 65485
    F17 = 65486
    F18 = 65487
    F19 = 65488
    F20 = 65489
    F21 = 65490
    F22 = 65491
    F23 = 65492
    F24 = 65493
    LSHIFT = 65505
    RSHIFT = 65506
    LCTRL = 65507
    RCTRL = 65508
    CAPSLOCK = 65509
    LMETA = 65511
    RMETA = 65512
    LALT = 65513
    RALT = 65514
    LWINDOWS = 65515
    RWINDOWS = 65516
    LCOMMAND = 65517
    RCOMMAND = 65518
    LOPTION = 65488
    ROPTION = 65489
    SPACE = 32
    EXCLAMATION = 33
    DOUBLEQUOTE = 34
    HASH = 35
    POUND = 35
    DOLLAR = 36
    PERCENT = 37
    AMPERSAND = 38
    APOSTROPHE = 39
    PARENLEFT = 40
    PARENRIGHT = 41
    ASTERISK = 42
    PLUS = 43
    COMMA = 44
    MINUS = 45
    PERIOD = 46
    SLASH = 47

    # Numbers on the main keyboard
    KEY_0 = 48
    KEY_1 = 49
    KEY_2 = 50
    KEY_3 = 51
    KEY_4 = 52
    KEY_5 = 53
    KEY_6 = 54
    KEY_7 = 55
    KEY_8 = 56
    KEY_9 = 57
    COLON = 58
    SEMICOLON = 59
    LESS = 60
    EQUAL = 61
    GREATER = 62
    QUESTION = 63
    AT = 64
    BRACKETLEFT = 91
    BACKSLASH = 92
    BRACKETRIGHT = 93
    ASCIICIRCUM = 94
    UNDERSCORE = 95
    GRAVE = 96
    QUOTELEFT = 96
    A = 97
    B = 98
    C = 99
    D = 100
    E = 101
    F = 102
    G = 103
    H = 104
    I = 105  # noqa: E741
    J = 106
    K = 107
    L = 108
    M = 109
    N = 110
    O = 111  # noqa: E741
    P = 112
    Q = 113
    R = 114
    S = 115
    T = 116
    U = 117
    V = 118
    W = 119
    X = 120
    Y = 121
    Z = 122
    BRACELEFT = 123
    BAR = 124
    BRACERIGHT = 125
    ASCIITILDE = 126



from .console import Console, Tile

from baggo import Color, colors, to_cp437


class SimpleConsole(Console):
    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

        num_tiles = width * height
        self.tiles = [Tile(0, colors.WHITE, colors.BLACK) for _ in range(num_tiles)]

    def at(self, x: int, y: int) -> Tile | None:
        index = self.try_index(x, y)
        if index is not None:
            return self.tiles[index]
        return None

    def clear(self, color: Color = colors.BLACK) -> None:
        for tile in self.tiles:
            tile.glyph = 32
            tile.foreground = colors.WHITE
            tile.background = color

        self.dirty = True

    def print(
        self,
        x: int,
        y: int,
        text: str,
        foreground: Color = colors.WHITE,
        background: Color = colors.BLACK,
    ) -> None:
        changed = False
        for char in text:
            cp = to_cp437(char)
            index = self.try_index(x, y)
            if index is not None:
                self.tiles[index].glyph = cp
                self.tiles[index].foreground = foreground
                self.tiles[index].background = background
                changed = True
            x += 1

        if changed:
            self.dirty = True

    def set(
        self,
        x: int,
        y: int,
        glyph: int,
        foreground: Color = colors.WHITE,
        background: Color = colors.BLACK,
    ) -> None:
        index = self.try_index(x, y)
        if index is not None:
            self.tiles[index].glyph = glyph
            self.tiles[index].foreground = foreground
            self.tiles[index].background = background
            self.dirty = True

    def index(self, x: int, y: int) -> int:
        return (self.height - 1 - y * self.width) + x

    def in_bounds(self, x: int, y: int) -> bool:
        return 0 <= x < self.width and 0 <= y < self.height

    def try_index(self, x: int, y: int) -> int | None:
        if self.in_bounds(x, y):
            return self.index(x, y)

        return None



import arcade
import array
from arcade.gl import BufferDescription
from pyglet.event import EVENT_HANDLE_STATE

from baggo import App, Keys
from baggo.terminal import Console, Terminal

from .font_arcade import FontArcade


# Vertex shader for instanced glyph rendering
VERTEX_SHADER = """
#version 330

// Per-vertex attributes (quad corners)
in vec2 in_vert;
in vec2 in_uv;

// Per-instance attributes
in vec2 in_pos;          // Screen position
in vec2 in_tex_offset;   // Texture atlas offset for this glyph
in vec4 in_fg_color;     // Foreground color (RGBA)
in vec4 in_bg_color;     // Background color (RGBA)

out vec2 v_uv;
out vec2 v_tex_offset;
out vec4 v_fg_color;
out vec4 v_bg_color;

uniform vec2 u_glyph_size;   // Glyph size in pixels
uniform vec2 u_screen_size;  // Screen size in pixels

void main() {
    // Calculate pixel position
    vec2 pixel_pos = in_pos + in_vert * u_glyph_size;

    // Convert to normalized device coordinates (-1 to 1)
    vec2 ndc = (pixel_pos / u_screen_size) * 2.0 - 1.0;
    ndc.y = -ndc.y; // Flip Y axis

    gl_Position = vec4(ndc, 0.0, 1.0);

    v_uv = in_uv;
    v_tex_offset = in_tex_offset;
    v_fg_color = in_fg_color;
    v_bg_color = in_bg_color;
}
"""

# Fragment shader
FRAGMENT_SHADER = """
#version 330

in vec2 v_uv;
in vec2 v_tex_offset;
in vec4 v_fg_color;
in vec4 v_bg_color;

out vec4 fragColor;

uniform sampler2D u_texture;
uniform vec2 u_glyph_uv_size;  // Size of one glyph in UV coords

void main() {
    // Calculate actual UV coordinates in the texture atlas
    vec2 atlas_uv = v_tex_offset + v_uv * u_glyph_uv_size;

    // Sample the texture (assuming white = foreground, black = background)
    float mask = texture(u_texture, atlas_uv).r;

    // Mix between background and foreground based on mask
    fragColor = mix(v_bg_color, v_fg_color, mask);
}
"""


class TerminalWindow(arcade.Window):
    def __init__(self, width: int, height: int, title: str, terminal: TerminalArcade):
        super().__init__(width, height, title)
        self.terminal = terminal

    def on_draw(self):
        self.terminal.on_draw()

    def on_update(self, delta_time: float):
        self.terminal.on_update(delta_time)

    def on_key_press(self, symbol: int, modifiers: int) -> EVENT_HANDLE_STATE:
        self.terminal.app.on_key_down(Keys(symbol), modifiers)

    def on_key_release(self, symbol: int, modifiers: int) -> EVENT_HANDLE_STATE:
        self.terminal.app.on_key_up(Keys(symbol), modifiers)


class TerminalArcade(Terminal):
    _font: FontArcade

    def __init__(
        self, width: int, height: int, title: str, console: Console, font: FontArcade
    ):
        self._width = width
        self._height = height
        self.console = console

        self.window = TerminalWindow(width, height, title, self)
        self.ctx = self.window.ctx

        self._font = font
        if not self._font.initialized:
            self._font.load()

        self.glyph_program = self.ctx.program(
            vertex_shader=VERTEX_SHADER, fragment_shader=FRAGMENT_SHADER
        )

        # Quad vertices
        # Format: x, y, u, v
        vertices = array.array(
            "f",
            [
                0.0,
                0.0,
                0.0,
                1.0,  # Bottom-left
                1.0,
                0.0,
                1.0,
                1.0,  # Bottom-right
                0.0,
                1.0,
                0.0,
                0.0,  # Top-left
                1.0,
                0.0,
                1.0,
                1.0,  # Bottom-right
                1.0,
                1.0,
                1.0,
                0.0,  # Top-right
                0.0,
                1.0,
                0.0,
                0.0,  # Top-left
            ],
        )

        self.quad_buffer = self.ctx.buffer(data=vertices)

        # Create instance buffer (will be updated each frame)
        # Format per instance: x, y, tex_offset_x, tex_offset_y, fg_r, fg_g, fg_b, fg_a, bg_r, bg_g, bg_b, bg_a
        max_instances = self.console.width * self.console.height
        self.instance_buffer = self.ctx.buffer(
            reserve=max_instances * 12 * 4
        )  # 12 floats per instance

        # Create geometry
        self.geometry = self.ctx.geometry(
            [
                BufferDescription(self.quad_buffer, "2f 2f", ["in_vert", "in_uv"]),
                BufferDescription(
                    self.instance_buffer,
                    "2f 2f 4f 4f",
                    ["in_pos", "in_tex_offset", "in_fg_color", "in_bg_color"],
                    instanced=True,
                ),
            ]
        )

        # Set uniforms
        self.glyph_program["u_glyph_size"] = (
            self._font.tile_width,
            self._font.tile_height,
        )
        self.glyph_program["u_screen_size"] = (width, height)
        self.glyph_program["u_glyph_uv_size"] = (
            1.0 / self._font.columns,
            1.0 / self._font.rows,
        )

        self.instance_count = 0

    def on_draw(self):
        self.window.clear()

        if self.instance_count > 0:
            # Bind font texture
            self._font.texture.use(0)
            self.glyph_program["u_texture"] = 0

            # Draw all glyphs in one call
            self.geometry.render(self.glyph_program, instances=self.instance_count)

    def on_update(self, delta_time: float):
        if self.app is not None:
            self.app.tick(delta_time)

        if self.console.dirty:
            # Build instance data
            instance_data = []

            for y in range(self.console.height):
                for x in range(self.console.width):
                    tile = self.console.at(x, y)
                    if tile is not None:
                        # Position
                        px = x * self._font.tile_width
                        py = y * self._font.tile_height

                        # Texture offset - calculate based on glyph position in atlas
                        glyph_col = tile.glyph % self._font.columns
                        glyph_row = tile.glyph // self._font.columns
                        glyph_x = glyph_col / self._font.columns
                        # Flip Y - texture origin is bottom-left, atlas layout is top-left
                        glyph_y = 1.0 - ((glyph_row + 1) / self._font.rows)

                        # Colors (normalize to 0-1 range)
                        fg_r, fg_g, fg_b = (
                            tile.foreground[0] / 255.0,
                            tile.foreground[1] / 255.0,
                            tile.foreground[2] / 255.0,
                        )
                        bg_r, bg_g, bg_b = (
                            tile.background[0] / 255.0,
                            tile.background[1] / 255.0,
                            tile.background[2] / 255.0,
                        )

                        instance_data.extend(
                            [
                                px,
                                py,  # Position
                                glyph_x,
                                glyph_y,  # Texture offset
                                fg_r,
                                fg_g,
                                fg_b,
                                1.0,  # Foreground color
                                bg_r,
                                bg_g,
                                bg_b,
                                1.0,  # Background color
                            ]
                        )

            # Update buffer
            if instance_data:
                self.instance_buffer.write(array.array("f", instance_data))
                self.instance_count = len(instance_data) // 12
            else:
                self.instance_count = 0

            self.console.dirty = False

    def register_app(self, app: App):
        self._app = app

    def run(self):
        arcade.run()
